/*
 * SPDX-License-Identifier: BSD-3-Clause
 * Copyright (c) 1996-2025, The SLICOT Team (original Fortran77 code)
 * Copyright (c) 2025, slicot.c contributors (C11 translation)
 */

#ifndef SLICOT_H
#define SLICOT_H

#include "slicot_types.h"

#include "slicot/ab.h"
#include "slicot/ag.h"
#include "slicot/bb.h"
#include "slicot/bd.h"
#include "slicot/de.h"
#include "slicot/df.h"
#include "slicot/dg.h"
#include "slicot/dk.h"
#include "slicot/fb.h"
#include "slicot/fd.h"
#include "slicot/ib.h"
#include "slicot/ma.h"
#include "slicot/mb01.h"
#include "slicot/mb02.h"
#include "slicot/mb03.h"
#include "slicot/mb04.h"
#include "slicot/mb05.h"
#include "slicot/mc.h"
#include "slicot/md.h"
#include "slicot/nf.h"
#include "slicot/sb.h"
#include "slicot/sg.h"
#include "slicot/tb.h"
#include "slicot/tc.h"
#include "slicot/td.h"
#include "slicot/tf.h"
#include "slicot/tg.h"
#include "slicot/ud.h"
#include "slicot/ue.h"
#include "slicot/lapack_aux.h"

#endif /* SLICOT_H */
